import { applyMiddleware as applyMiddlewareWithRedux, Middleware, Store } from 'redux';
import { storeKey } from '../constants';
/**
 * ## Description
 * Apply middlewares for Redux.
 *
 * ## Example
 *
 * ```ts
 * import logger from 'redux-logger';
 *
 * @injectable()
 * class Foo {}
 *
 * const app = createApp({
 *   modules: [applyMiddleware(logger)],
 *   main: Foo,
 *   render: () => {},
 * });
 * ```
 *
 * @param args middlewares for Redux
 */
declare const applyMiddleware: (...args: Middleware[]) => {
    new (): {
        readonly [storeKey]?: any;
        enhancer: any;
        preloadedStateHandler?(preloadedState: applyMiddlewareWithRedux<any>): applyMiddlewareWithRedux<any>;
        middleware?: Middleware;
        afterCreateStore?(store: Store): Store;
        beforeCombineRootReducers?(reducers: applyMiddlewareWithRedux): applyMiddlewareWithRedux;
        afterCombineRootReducers?(rootReducer: applyMiddlewareWithRedux): applyMiddlewareWithRedux;
        provider?: import("react").FunctionComponent<any>;
    };
};
export { applyMiddleware };
//# sourceMappingURL=applyMiddleware.d.ts.map